/** This file is part of Reactor.
 *  Copyright (c) 2021-2023 Kedron Holdings LLC, All Rights Reserved.
 *  Reactor is not public domain or open source. Distribution or derivative works are expressly prohibited.
 */

const version = 23063;

const fs = require('fs');
const path = require('path');

const Configuration = require("server/lib/Configuration");
const Capabilities = require("server/lib/Capabilities");
const Controller = require("server/lib/Controller").requires( 22280 );
const Logger = require("server/lib/Logger");
const Timer = require("server/lib/Timer");

const Entity = require("server/lib/Entity").requires( 22314 );
const Group = require("server/lib/entities/Group");

const util = require("server/lib/util");
const isUndef = util.isUndef;
const coalesce = util.coalesce;

const lexp = require('common/lexp');

Logger.getLogger( 'ZWaveJSController', 'Controller' ).always( "Module ZWaveJSController v%1", version );

/* ??? These values, and other similar throughout, need to be config */
const MAX_FAIL = 3;         /* Maximum number of connection attempts before signalling offline */
const ERROR_DELAY = 5000;   /* Delay between connection attempts */
const PING_INTERVAL = 90000; /* ZWaveJS seems to ping every 56s, so this is a safe value */
const DEFAULT_TASK_TIMEOUT = 20000;

const TRUE_RE = /^(true|yes|on)$/i;

const NODE_STATUS_TEXT = [ 'unknown', 'asleep', 'awake', 'dead', 'alive' ];
const NODE_ASLEEP = 1;
const NODE_AWAKE = 2;
const NODE_DEAD = 3;
const NODE_ALIVE = 4;

const dumpdir = Configuration.getConfig( "reactor.logsdir" );
const confdir = Configuration.getConfig( "reactor.confdir" );

let last_req_id = Date.now();
let impl = false, CC = {};
let expr_ce = {};

const BLANK = function( s ) {
    return ( "undefined" === typeof s || null === s ) ? "" : String(s);
};

module.exports = class ZWaveJSController extends Controller {
    constructor( struct, id, config ) {
        super( struct, id, config );

        this.stopping = false;
        this.retries = 0;
        this.pending = {};

        this.nodes = {};
        this.poll_queue = [];
        this.poll_timer = new Timer( this.toString() + "#poller", this._poll_handler.bind( this ) );

        this.config.source = this.config.source || "ws://127.0.0.1:3000/";
        this.config.poll_interval = coalesce( parseInt( this.config.poll_interval ), 3600 );
        this.config.poll_frequency = Math.max( 10, coalesce( parseInt( this.config.poll_frequency ), 20 ) ) * 1000;
        try {
            let streams = this.log.getStreams();
            if ( ! ( streams && streams.find( el => 'file' === el.type ) ) ) {
                this.log.addStream( 'file', 999, { name: this.getID() + '.log' } );
                this.log.notice( "%1 added file logger to %2", this, this.getID() + ".log" );
            }
        } catch ( err ) {
            this.log.exception( err );
        }
        if ( false && this.log.getLevel() < 5 ) {
            this.log.setLevel( 5 );
        }
    }

    /** Start the controller.
     */
    async start() {
        this.stopping = false;
        if ( !impl ) {
            await this._load_implementation();
        }
        this.cookie = this.getStructure().get_host_public_key();
        await this.run();
        return this;
    }

    async stop() {
        this.stopping = true;

        this.poll_timer.cancel();
        this.poll_queue = [];
        this.nodes = {};

        return await super.stop();
    }

    async run() {
        if ( this.stopping ) {
            return;
        }

        if ( this.config.load_static_data ) {
            this.log.notice( "%1 loading static device inventory from %2", this, this.config.load_static_data );
            let data = JSON.parse( fs.readFileSync( this.config.load_static_data ) );
            if ( data.hostkey && data.data ) {
                data = data.data;
            }
            this.purgeAllEntities();
            /* This little horror show is because the Entity's send of its delete notification is asynchronous,
               and we end up racing the delete with the creation of a new entity if we don't wait. Someday, that
               needs to be fixed the right way. See Entity.destroy()
             */
            setTimeout( () => {
                this._process_init_data( data );
                this.online();
                this.log.warn( "%1 controller in static data mode with %2", this, this.config.load_static_data );
                this.stopping = true;
            }, 2000 );
            return;
        }

        if ( ! this.connected ) {
            this.log.info( "%1 connecting to %2", this, this.config.source );
            this.ws_open( this.config.source, {
                    pingInterval: coalesce( parseInt( this.config.ping_interval ), 60000 ),
                    connectTimeout: coalesce( parseInt( this.config.connect_timeout ), 15000 ),
                    ignoreCertificates: !! this.config.ignore_cert
                }, this.config.wsoptions || {} ).then( async ( hello ) => {
                this.log.notice( "%1 connected; performing initial inventory", this );
                // Ref: https://github.com/zwave-js/zwave-js-server
                // On connect, we immediately get a "version" message type we should check: https://github.com/zwave-js/zwave-js-server#schema-version
                await this._send_tracked_message( { command: "set_api_schema", schemaVersion: 14 } ); // or maybe 10 ???
                if ( this.config.listen_zwavejs_logs ) {
                    await this._send_tracked_message( { command: "driver.start_listening_logs" } );
                }

                this._send_tracked_message( { command: "start_listening" } ).then( ( reply ) => {
                    let data = reply.result;
                    this.log.debug( 9, "%1 got initialization data", this );
                    try {
                        fs.writeFileSync( path.join( dumpdir, this.getID() + "_devices_initial.json" ), JSON.stringify( data, null, 4 ) );
                    } catch ( err ) {
                        this.log.err( "%1 failed to write initialization capture: %2", this, err );
                    }

                    this._process_init_data( data ).then( () => {
                        this.retries = 0;
                        this.online();

                        if ( this.authorized ) { this.log.notice( "%1 checking poller", this ); }
                        this._poll_begin();

                        if ( true !== this.config.data_collection_opt_out ) {
                            // Delay this two minutes to give interviews time to complete.
                            const self = this;
                            setTimeout( () => {
                                const os = require( 'os' );
                                self.fetchJSON( 'https://reactor.toggledbits.com/supportapi/zwjs_collect.php',
                                    {
                                        method: 'POST',
                                        timeout: 60000,
                                        headers: {
                                            accept: 'application/json',
                                            'content-type': 'application/json'
                                        },
                                        body: JSON.stringify(
                                            {
                                                format: 1,
                                                hostkey: self.getStructure().get_host_public_key(),
                                                controller_version: version,
                                                authorized: self.authorized,
                                                platform: os.platform(),
                                                arch: os.arch(),
                                                docker: fs.existsSync( '/opt/reactor/.reactor-docker' ),
                                                data: data
                                            },
                                            ( key, val ) => key.startsWith( '__' ) ? undefined : val
                                        )
                                }).catch( err => {
                                    self.log.exception( err );
                                });
                            }, 120000 );
                        }
                    });
                }).catch( err => {
                    throw err;
                });
            }).catch ( err => {
                this.log.notice( "%1 connection to %2 failed; will retry", this, this.config.source );
                this.ws_terminate();
                if ( ++this.retries >= 3 ) {
                    this.offline();
                }
                this.stopTimer();
                let delay = Math.min( 120000, Math.max( 0, this.retries - 12 ) * 10000 + 10000 );
                this.log.info( "%1 retry %3 delay %2", this, delay, this.retries );
                this.startDelay( delay );
            });
        }
    }

    _process_init_data( data ) {
        try {
            // const buff = Buffer.from( this.config.auth || "", "base64" );
            // const dec_res = this.getStructure().unpack_tb( buff ).toString();
            // this.authorized = dec_res === "start_listening";
            this.authorized = version;
        } catch ( err ) {
            this.log.exception( err );
        }

        const t0 = Date.now();
        return this._register_nodes( data.state.nodes ).then( () => {
            this.log.info( "%1 Node registration complete, %2ms", this, Date.now() - t0 );

            const zw = this.findEntity( `${data.state.controller.ownNodeId || 1}-0` );
            if ( zw ) {
                zw.extendCapability( 'zwave_network' );
                zw.setAttributes({
                    'is_secondary': coalesce( data.state.controller.isSecondary ),
                    'library_version': coalesce( data.state.controller.libraryVersion ),
                    'is_suc': coalesce( data.state.controller.isStaticUpdateController ),
                    'is_sis_present': coalesce( data.state.controller.isSISPresent ),
                    'suc_node': coalesce( data.state.controller.sucNodeId ),
                    'state': true
                }, `zwave_network` );
            }
        }).catch( err => {
            this.log.exception( err );
        });
    }

    // Ref: https://github.com/zwave-js/zwave-js-server#zwave-js-events
    ws_message( msg ) {
        let data = JSON.parse( msg );
        if ( data.type ) {
            let impl = this[ "_wsmessage_" + data.type ];
            if ( impl ) {
                try {
                    impl.call( this, data );
                } catch ( err ) {
                    this.log.err( "%1 _wsmessage_%2 threw untrapped exception: %3", this, data.type, err );
                    this.log.exception( err );
                }
            } else {
                this.log.debug( 5, "%1 ignoring message type %2 (no handler)", this, data.type );
                this.log.debug( 8, "data %1", data );
            }
        }
    }

    ws_closing() {
        if ( ! this.stopping ) {
            this.log.notice( "%1 connection to Z-WaveJS closing (%2); will attempt reconnect", this, this.config.source );
        }
        this.poll_timer.cancel();
        this.stopTimer();
        this.startDelay( 2000 );
    }

    _next_req_id() {
        let n = Date.now();
        if ( n <= last_req_id ) {
            n = last_req_id + 1;
        }
        last_req_id = n;
        return n;
    }

    /** Returns a Promise that sends a request to Hass and resolves when it gets
     *  the matching reply. The reply can time out, and the Promise rejects.
     */
    _send_tracked_message( req, timeout ) {
        const self = this;
        timeout = timeout || DEFAULT_TASK_TIMEOUT;
        let slot = {
            req_id: this._next_req_id(),
            expires: Date.now() + timeout,
            resolve: false,
            reject: false,
            type: req.command
        };
        slot.promise = new util.TimedPromise( (resolve,reject) => {
            slot.resolve = resolve;
            slot.reject = reject;
            req.messageId = slot.req_id;
            self.log.debug( 5, "%1 sending #%2: %3", this, slot.req_id, req );
            self.ws_send( JSON.stringify( req ) );
        }, timeout ).catch( err => {
            self.log.err("%1 request %2 (%4) failed: %3", self, slot.req_id, err, slot.type);
            slot.reject( err );
            delete self.pending[String(slot.req_id)];
        });
        this.pending[String(slot.req_id)] = slot;
        return slot.promise;
    }

    /** Greeting from server on connect */
    _wsmessage_version( event ) {
        this.log.info( "%1 greeting from server: %2", this, event );
    }

    /** Called for "result" type messages. If associated with a tracked request, that Promise will
     *  resolve or reject based on the returned success flag. Otherwise, the message is logged (if
     *  level is high enough) and the message discarded.
     */
    _wsmessage_result( event ) {
        this.log.debug( 8, "%1 result response for message %2 type %3 success %4",
            this, event.messageId, event.type, event.success );
        this.log.debug( 9, "%1 full result reply %2", this, event );
        let slot = this.pending[String(event.messageId)];
        if ( slot ) {
            delete this.pending[String(event.messageId)];
            this.log.debug( 8, "%1 resolving pending #%1", this, event.messageId );
            // ??? They all seem to succeed?
            if ( event.success && false !== this.authorized ) {
                this.log.debug( 5, "%1 request %2 (%3) success notification", this, slot.req_id, slot.type );
                slot.resolve( event );
            } else {
                // Ref: https://github.com/zwave-js/zwave-js-server#errors
                this.log.debug( 5, "%1 command result error: %2", this, event );
                let err;
                if ( "zwave_error" === event.errorCode ) {
                    err = new Error( `Z-Wave error ${event.zwaveErrorCode}: ${event.zwaveErrorMessage}` );
                } else {
                    err = new Error( event.errorCode );
                }
                slot.reject( err );
            }
        } else {
            this.log.debug( 6, "%1 can't resolve request %2 in reply, slot not found, maybe already timed out", this,
                event.messageId );
        }
    }

    /** Called for "event" messages, which mean a lot of things. */
    async _wsmessage_event( event ) {
        // Ref: https://zwave-js.github.io/node-zwave-js/#/api/node?id=zwavenode-events
        switch ( event.event.source ) {
            case "node":
                {
                    if ( ! this.nodes[ event.event.nodeId ] ) {
                        this.log.info( "%1 received event for unknown node %2; querying...", this, event.event.nodeId );
                        this._get_add_node( event.event.nodeId );
                        return;
                    }
                    const entity = this.findEntity( `${event.event.nodeId}-${event.event.endpoint || 0}` );
                    if ( ! entity ) {
                        this.log.debug( 7, "%1 received %2 event for node %3, no entity (ignored)", this,
                            event.event.event, event.event.nodeId );
                        return;
                    }
                    this.log.debug( 5, "%1 handling node event %2 entity %3", this, event.event.event, entity );
                    /* Node Monitoring (writes all node events) */
                    if ( this.config.monitor_nodes && Array.isArray( this.config.monitor_nodes ) &&
                            this.config.monitor_nodes.find( n => n === event.event.nodeId ) ) {
                        try {
                            fs.writeFileSync(
                                path.join( dumpdir, `${this.getID()}_${event.event.nodeId}_monitor.json` ),
                                JSON.stringify( event.event, null, 4 ), { flag: "a" } );
                        } catch ( err ) {
                            this.log.exception( err );
                        }
                    }
                    switch ( event.event.event ) {
                        case "value updated":
                        case "value notification":
                            await this._update_value( event.event );
                            break;

                        case "value added":
                            {
                                // { "type": "event", "event": { "source": "node", "event": "value added", "nodeId": 15, "args": { "commandClassName": "Scene Actuator Configuration", "commandClass": 44, "property": "level", "propertyKey": 1, "newValue": 1, "propertyName": "level", "propertyKeyName": "1" } } }
                                const N = await this.nodes[ event.event.nodeId ];
                                if ( N ) {
                                    const endp = N.endpoints[ event.event.endpoint || 0 ];
                                    if ( endp ) {
                                        const vid = `${event.event.args.commandClass}:${event.event.args.property}:${BLANK(event.event.args.propertyKey)}`;
                                        endp.__values.push( event.event.args );
                                        endp.__valueIndex[ vid ] = event.event.args;
                                        /* ??? get metadata, fix up values/valueIndex entries */

                                        /* ??? set capabilities/attributes */

                                        let fullname = `x_zwave_values.`;
                                        fullname += String( event.event.args.commandClassName || event.event.args.commandClass ).replace( /[^A-Z0-9_]/ig, "_" );
                                        fullname += "_" + String( event.event.args.propertyName ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                                        if ( "undefined" !== typeof event.event.args.propertyKey && null !== event.event.args.propertyKey ) {
                                            fullname += "_" + String( event.event.args.propertyKey ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                                        }
                                        this.log.debug( 5, "%1 setting %2.%3 to %4", this, entity, fullname, event.event.args.newValue );
                                    }
                                }
                            }
                            break;

                        case "ready":
                            {
                                /* Device has come ready; refresh to make sure we have all CCs and mappings */
                                this.log.info( "%1 node %2 ready; refreshing...", this, event.event.nodeId );
                                this._get_add_node( event.event.nodeId );
                                // ??? send queued commands?
                            }
                            break;

                        case "dead":
                        case "alive":
                            {
                                const alive = event.event.event === "alive";
                                entity.setAttributes( {
                                    "failed": ! alive,
                                    "status": alive ? impl.node_status.alive : impl.node_status.dead,
                                    "status_text": coalesce( event.event.event )
                                }, "zwave_device" );
                            }
                            break;

                        case "interview failed":
                            {
                                this.log.notice( "%1 node %2 interview failed; marking dead", this, event.event.nodeId );
                                entity.setAttributes( {
                                    "failed": true,
                                    "status": impl.node_status.dead,
                                    "status_text": "dead"
                                }, "zwave_device" );
                            }
                            break;

                        case "wake up":
                            {
                                entity.setAttributes( {
                                    "zwave_device.last_wakeup": Date.now(),
                                    "zwave_device.status": impl.node_status.awake,
                                    "zwave_device.status_text": 'awake'
                                });

                                // ??? send commands queued while sleeping?
                            }
                            break;

                        case "sleep":
                            {
                                entity.setAttributes( {
                                    "zwave_device.status": impl.node_status.asleep,
                                    "zwave_device.status_text": 'asleep'
                                });
                            }
                            break;

                        case "statistics updated":
                            // { "type": "event", "event": { "source": "node", "event": "statistics updated", "nodeId": 15, "statistics": { "commandsTX": 63, "commandsRX": 4590, "commandsDroppedRX": 0, "commandsDroppedTX": 0, "timeoutResponse": 0 } } }
                            this.log.debug( 7, "%1 received statistics for %2: %3", this, entity, event.event.statistics );
                            break;

                        // metadata updated
                        // notification
                        default:
                            this.log.debug( 5, "%1 unhandled node event %2", this, event.event.event );
                            this.log.debug( 6, "%1 event %2", this, event );
                    }
                }
                break;

            case "controller":
                {
                    this.log.debug( 7, "%1 handling controller event %2", this, event.event.event );
                    switch ( event.event.event ) {
                        case "statistics updated":
                            // { "type": "event", "event": { "source": "controller", "event": "statistics updated", "statistics": { "messagesTX": 9, "messagesRX": 135, "messagesDroppedRX": 0, "NAK": 0, "CAN": 0, "timeoutACK": 0, "timeoutResponse": 0, "timeoutCallback": 0, "messagesDroppedTX": 0 } } }
                            break;

                        case "node added":
                            /* Not sure what the message looks like; query the whole state. */
                            this.log.info( "%1 received node added notification for %2; registering...", this, event.event.nodeId );
                            this._get_add_node( event.event.nodeId );
                            break;

                        case "node removed":
                            /* Do nothing. */
                            break;

                        // inclusion,exclusion started,failed,stopped
                        // heal network progress,done
                        default:
                            this.log.debug( 5, "%1 unhandled controller event %2", this, event.event.event );
                            this.log.debug( 6, "%1 event %2", this, event );
                    }
                }
                break;

            case "driver":
                {
                    if ( "logging" === event.event.event ) {
                        this.log.raw( event.event.formattedMessage );
                    } else {
                        this.log.debug( 5, "%1 unhandled driver event %2", this, event.event.event );
                    }
                }
                break;

            default:
                this.log.debug( 5, "%1 unhandled event source %2", this, event.event.source );
                this.log.debug( 6, "%1 event %2", this, event );
        }
    }

    _get_add_node( nodeId ) {
        return this._send_tracked_message( { command: "node.get_state", nodeId: nodeId } ).then( ev => {
            try {
                this.nodes[ nodeId ] = this._register_node( ev.result.state );
            } catch ( err ) {
                this.log.err( "%1 exception in _get_add_node(): %2", this, err );
                this.log.exception( err );
                this.log.debug( 5, "ev=%1", ev );
            }
        }).catch ( err => {
            throw err;
        });
    }

    async _load_implementation() {
        impl = await this.loadBaseImplementationData( 'zwavejs', __dirname );

        let r = util.loadData( [ "." ], "local_zwavejs_devices", confdir );
        if ( r ) {
            this.log.notice("%1 applying local implementation overrides", this );
            this.log.logp( 8, this, '_load_implmap', "local implementation %1", r.implementation );
            impl.implementation = util.combine( impl.implementation || {}, r.implementation || {} );
            this.log.logp( 8, this, '_load_implmap', 'local mapping contains %1 entries', ( r.mapping || [] ).length );
            if ( ( r.mapping || [] ).length > 0 ) {
                impl.mapping = r.mapping.concat( impl.mapping || [] );
            }
            if ( null !== r.entities && "object" === typeof r.entities ) {
                impl.entities = r.entities;
            }
            r = undefined;
        }
        impl.entities = impl.entities || {};

        impl.manufacturer = impl.manufacturer || {};

        CC = new Map();
        for ( let ccid of Object.keys( impl.command_class ) ) {
            if ( isNaN( ccid ) ) {
                this.log.err( "%1 zwave_devices.yaml command_class %2 not integer (should be)", this, ccid );
            } else if ( impl.command_class[ ccid ].name ) {
                CC.set( parseInt( ccid ), impl.command_class[ ccid ].name );
                CC.set( impl.command_class[ ccid ].name, parseInt( ccid ) );
            } else {
                this.log.err( "%1 zwave_devices.yaml command_class 0x%2 missing name", this, parseInt( ccid ).toString(16) );
            }
        }
    }

    _value_match( entry, value ) {
        if ( null === entry.match || "object" !== typeof entry.match ) {
            this.log.err( "%1 malformed match clause in %2", this, entry );
            return false;
        }
        for ( const crit of Object.keys( entry.match || {} ) ) {
            let val = util.deref( value, crit );
            if ( val !== entry.match[ crit ] ) {
                return false;
            }
        }
        return entry;
    }

    _endpoint_match( entry, endpoint ) {
        if ( null === entry.endpoint_match || "object" !== typeof entry.endpoint_match ) {
            this.log.err( "%1 malformed endpoint_match clause in %2", this, entry );
            return false;
        }
        for ( const crit of Object.keys( entry.endpoint_match || {} ) ) {
            let val = util.deref( endpoint, crit );
            if ( val !== entry.endpoint_match[ crit ] ) {
                return false;
            }
        }
        return entry;
    }

    _apply_value( entity, fullname, endpoint, impl, ctx ) {
        this.log.debug( 5, "%1 _apply_value entity %2 %3", this, entity, fullname );
        let newval = null;
        if ( "object" !== typeof impl || null === impl ) {
            newval = impl;
        } else {
            let vv = null;
            if ( impl.value ) {
                newval = impl.value;
            } else {
                if ( impl.valueId ) {
                    /* From different value on this endpoint, any CC */
                    vv = endpoint.__valueIndex[ impl.valueId ];
                    this.log.debug( 6, "%1 handling %2 %3 by %4 with %5", this, entity, fullname, impl, vv );
                    /* deref() as of 22036 is safe for undefined object, and has default return arg */
                    newval = util.deref( vv, impl.source || 'value', undefined );
                    if ( isUndef( newval ) ) {
                        if ( impl.skip_missing ) {
                            this.log.debug( 5, "%1 value missing from %2:%3, skipped", this, endpoint.index, impl.valueId );
                            return;
                        } else {
                            this.log.info( "%1 node %2-%3 no value %4 found for %5",
                                this, entity.getAttribute( 'zwave_device.node_id' ), endpoint.index, impl.valueId,
                                fullname );
                        }
                        newval = null;
                    }
                } else if ( ! impl.expr ) {
                    this.log.err( "%1 no source for %2 in %3", this, fullname, impl );
                }
            }

            if ( impl.expr ) {
                if ( isUndef( expr_ce[ impl.expr ] ) ) {
                    try {
                        expr_ce[ impl.expr ] = lexp.compile( impl.expr );
                    } catch ( err ) {
                        this.log.err( "%1 can't compile expression for %2, %3 = %4", this, fullname, impl.expr, err );
                        expr_ce[ impl.expr ] = false;
                    }
                }
                if ( expr_ce[ impl.expr ] ) {
                    try {
                        lexp.define_var( ctx, 'value', newval );
                        lexp.define_var( ctx, 'current_value', coalesce( vv ) );
                        lexp.define_var( ctx, 'entity', entity.extract() );
                        newval = lexp.run( expr_ce[ impl.expr ], ctx );
                    } catch ( err ) {
                        this.log.err( "%1 failed evaluation for %2, %3 = %4", this, fullname, impl.expr, err );
                        newval = null;
                    }
                }
            }
            if ( impl.map && ! isUndef( impl.map[ newval ] ) ) {
                /* ??? map_default? */
                newval = impl.map[ newval ];
            }
            if ( ( newval === null || isUndef( newval ) ) && ! isUndef( impl.default ) ) {
                newval = impl.default;
            }
        }
        if ( isUndef( newval ) ) {
            newval = null;
        }
        try {
            this.log.debug( 5, "%1 _apply_value %2=%3", this, fullname, newval );
            entity.setAttribute( fullname, newval );
        } catch ( err ) {
            this.log.exception( err );
        }
    }

    /** Given a node value update event, process it */
    async _update_value( eventdata ) {
        const k = `${eventdata.args.commandClass}:${eventdata.args.property}:${BLANK(eventdata.args.propertyKey)}`;
        const ek = `${eventdata.args.endpoint || 0}:${k}`;

        this.log.debug( 5, "%1 update node %2 value %3 data %4", this, eventdata.nodeId, ek, eventdata );

        const N = await this.nodes[ eventdata.nodeId ];
        if ( ! N ) {
            this.log.warn( "%1 update for unknown node %2", this, eventdata.nodeId );
            return;
        }

        const endpoint = N.endpoints[ eventdata.args.endpoint || 0 ];
        const value = endpoint.__valueIndex[ k ];
        if ( ! value ) {
            this.log.notice( "%1 update value %2 not on node %3 endpoint %4", this, k, N.nodeId, endpoint.index );
            return;
        }

        const newval = coalesce( ( "value notification" === eventdata.event ) ? eventdata.args.value : eventdata.args.newValue );
        const attributes = Object.keys( value.__endp || {} );
        this.log.debug( 5, "%1 updating attributes for node %2 value %3=%4: %5", this, eventdata.nodeId, ek,
            newval, attributes );
        endpoint.__valueIndex[ k ].value = newval;

        let ctx = lexp.get_context();
        lexp.define_var( ctx, 'system', this.getSystemEntity().extract() );
        lexp.define_var( ctx, 'values', endpoint.__valueIndex );
        lexp.define_var( ctx, 'node', N );
        lexp.define_var( ctx, 'endpoint', endpoint );

        let targets = new Set();
        try {
            for ( let fullname of attributes ) {
                const targetInfo = value.__endp[ fullname ];
                this.log.debug( 5, "%1 updating attribute %2 with %3", this, fullname, targetInfo );
                const entity = this.findEntity( targetInfo.entity );
                if ( ! entity ) {
                    this.log.warn( "%1 update %2 from %3 for unknown entity %5", this, fullname, ek, eventdata.nodeId );
                    continue;
                }
                entity.deferNotifies( true );
                entity.markUpdate();  /* No assurance setAttribute() will be called, so force mark */
                targets.add( entity );

                this._apply_value( entity, fullname, endpoint, targetInfo.impl, ctx );
            }

            let endpoint_entity = this.findEntity( `${N.nodeId}-${endpoint.index}` );
            if ( endpoint_entity ) {
                endpoint_entity.deferNotifies( true );
                /* No need to markUpdate() here, as setAttribute() does it for us when called. */
                targets.add( endpoint_entity );
                let n = `x_zwave_values.`;
                n += String( eventdata.args.commandClassName || eventdata.args.commandClass ).replace( /[^A-Z0-9_]/ig, "_" );
                n += "_" + String( eventdata.args.propertyName ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                if ( "undefined" !== typeof eventdata.args.propertyKey && null !== eventdata.args.propertyKey ) {
                    n += "_" + String( eventdata.args.propertyKey ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                }
                this.log.debug( 5, "%1 setting %2.%3 to %4", this, endpoint_entity, n, newval );
                endpoint_entity.setAttribute( n, newval );
            }
        } finally {
            for ( let target of targets.values() ) {
                try {
                    target.deferNotifies( false );
                } catch ( err ) {
                    this.log.err( "%1 release %2: %3", this, target, err );
                    this.log.exception( err );
                }
            }
        }
    }

    _match_devicetype( node ) {
        for ( let crit of impl.devicetype ) {
            if ( crit.expr ) {
                this.log.err( "%1 haven't implemented expression matching yet" );
            } else if ( crit.match ) {
                this.log.debug( 7, "%1 matching %2", this, crit );
                let matched = false;  /* Guilty until proven innocent */
                let patterns = crit.match.split( /;/ );
                for ( let pat of patterns ) {
                    let m = pat.split( /=/ );
                    let val = util.deref( node, m[0] );
                    if ( 1 === m.length ) {
                        /* No = means no value to match, just existence check, and we failed. */
                        matched = ! isUndef( val );
                    } else {
                        this.log.debug( 7, "%1 matching %2", this, m );
                        let matchval = m[1].trim();
                        if ( matchval.startsWith( '/' ) ) {
                            /* Regular expression match */
                            let ix = matchval.lastIndexOf( '/' );
                            let rep = matchval.substring( 1, ix );
                            let flags = matchval.substring( ix+1 );
                            rep = new RegExp( rep, flags );
                            if ( ! rep.test( val ) ) {
                                matched = false;
                                break;
                            }
                        } else if ( "boolean" === typeof val ) {
                            /* Numeric match if node value is numeric */
                            matched = val === TRUE_RE.test( matchval );
                        } else if ( "number" === typeof val ) {
                            /* Numeric match if node value is numeric */
                            matched = val === parseInt( matchval ) ;
                        } else {
                            matched = val === matchval;
                        }
                    }
                    if ( ! matched ) {
                        /* Didn't match; all must match. We're done here. */
                        break;
                    }
                }
                if ( matched ) {
                    return crit;
                }
            }
        }
        return undefined;
    }

    _register_nodes( nodes ) {
        /* "Compile" the node data into an indexable form. */
        let oldNodes = new Set( Object.keys( this.nodes || {}  ) );
        let promises = [];
        for ( let node of nodes ) {

            oldNodes.delete( String( node.nodeId ) );

// if ( this.config.toggledbits && 22 !== node.nodeId && 1 !== node.nodeId ) continue; // FILTER ONE NODE

            this.nodes[ node.nodeId ] = this._register_node( node );
            this.nodes[ node.nodeId ].__zwave_node = node.nodeId;
            promises.push( this.nodes[ node.nodeId ] );
        }

        try {
            /* Remove nodes that no longer exist. */
            for ( let nodeId of oldNodes.values() ) {
                this._remove_node( nodeId );
            }

            /* Remove entities for nodes that no longer exist. */
            for ( let entity of Object.values( this.getEntities() ) ) {
                if ( entity.hasCapability( 'zwave_device' ) ) {
                    const nid = entity.getAttribute( 'zwave_device.node_id' );
                    if ( ! this.nodes[ nid ] ) {
                        this.log.notice( "%1 removing %2, node %3 no longer exists", this, entity, nid );
                        entity.markDead( true );
                        if ( 0 === entity.getAttribute( 'zwave_device.endpoint' ) ) {
                            this.sendWarning(
                                _T( "Controller {0} node {1} ({2}) no longer exists.", this.getID(), nid, e.getName() )
                            );
                        }
                    }
                }
            }

            this.purgeDeadEntities( 1440 * 3 ); /* Default 72 hours purges dead entities automatically */
        } catch ( err ) {
            this.log.exception( err );
        }

        const self = this;
        return Promise.allSettled( promises ).then( results => {
            results.forEach( ( result, ix ) => {
                if ( "rejected" === result.status ) {
                    self.log.err( "%1 node %2 setup failed: %3", self, promises[ix]?.__zwave_node, result.reason );
                    self.log.exception( result.reason );
                }
            });
        });
    }

    async _register_node( N ) {
        this.log.debug( 5, "%1 registering node %2 %3 (%4)", this, N.nodeId,
            N.name || N.label, N.deviceClass?.generic?.label );

        /* Node Monitoring */
        if ( this.config.monitor_nodes && Array.isArray( this.config.monitor_nodes ) &&
                this.config.monitor_nodes.find( n => n === N.nodeId ) ) {
            try {
                fs.writeFileSync( path.join( dumpdir, `${this.getID()}_${N.nodeId}_monitor.json` ),
                    JSON.stringify( N, null, 4 ) );
            } catch ( err ) {
                this.log.exception( err );
            }
        }

        /* Try to save a little memory by removing big things we don't use/need */
        if ( N.deviceConfig ) {
            /* Metadata has instructions for inclusion, exclusion, reset, etc. */
            delete N.deviceConfig.metadata;
        }

        /* Remap endpoints to indexable form */
        N.endpoints = N.endpoints || [];
        for ( let endpoint of N.endpoints || [] ) {
            N.endpoints[ endpoint.index ].__values = [];
            N.endpoints[ endpoint.index ].__valueIndex = {};
            N.endpoints[ endpoint.index ].__commandClasses = new Set();
        }

        /* Remap values to indexable form per-endpoint. */
        N.__valueIndex = {};
        N.values.forEach( ( value, index ) => {
            /* This valueId is unique to each endpoint (but maybe not unique in node) */
            const valueId = value.commandClass + ":" + value.property + ":" + BLANK(value.propertyKey);
            N.__valueIndex[ `${value.endpoint}:${valueId}` ] = value;
            const endpoint = N.endpoints[ value.endpoint ];
            endpoint.__commandClasses.add( value.commandClass );
            endpoint.__values.push( value );
            endpoint.__valueIndex[ valueId ] = value;

            if ( ! CC.has( value.commandClass ) ) {
                this.log.warn( "%1 node %2-%3 has commandClass %4 not defined in zwavejs_devices.yaml",
                    this, N.nodeId, value.endpoint, value.commandClass );
            } else {
                const name = value.commandClass;
                if ( CC.get( name ) !== value.commandClassName ) {
                    this.log.warn( "%1 node %2-%3 has commandClass %4 with different name, yaml=%5, endp=%6",
                        this, N.nodeId, value.endpoint, value.commandClass, name, value.commandClassName );
                }
            }
        });

        let ctx = lexp.get_context( { system: this.getSystemEntity().extract() } );

        const node = N;

        /* Try to load manufacturer-specific data */
        if ( N.manufacturerId && isUndef( impl.manufacturer[ N.manufacturerId ] ) ) {
            this.log.debug( 5, "%1 loading manufacturer-specific data for %2", this, N.manufacturerId );
            let source = "package";
            try {
                let mfg = util.loadData( ['manufacturer'], String(N.manufacturerId), __dirname ) || null;
                const zipper = this.getZipper( 'zwavejs' );
                if ( zipper ) {
                    try {
                        const zm = await zipper.getJSONEntry( `manufacturer/${N.manufacturerId}.json` );
                        if ( !mfg || !mfg.version || !mfg.revision || zm.version >= mfg.version ||
                            ( zm.version === mfg.version && zm.revision > mfg.revision ) ) {
                            this.log.debug( 5, "%1 manufacturer %2 ZIP %3.%4 overrides package %5.%6", this,
                                N.manufacturerId, zm.version, zm.revision, mfg?.version, mfg?.revision );
                            mfg = zm;
                            source = "ZIP";
                        }
                    } catch ( err ) {
                        this.log.warn( "%1: no manufacturer %2 data: %3", zipper.getName(), N.manufacturerId, err );
                    }
                }
                const lnam = `zavejs_${N.manufacturerId}`;
                const lmfg = util.loadData( confdir, lnam, __dirname );
                if ( lmfg ) {
                    if ( lmfg.version > mfg.version || ( lmfg.version === mfg.version && lmfg.revision > mfg.revision ) ) {
                        mfg = impl.manufacturer[ N.manufacturerId ] = lmfg;
                        source = path.resolve( confdir, lnam );
                        this.log.debug( 5, "%1 selected local manufacturer %2 data in %3", this, N.manufacturerId,
                            path.resolve( confdir, lnam ) );
                    } else {
                        this.log.warn( "%1 local manufacturer data %2 ignored; %3 more recent", this,
                            path.resolve( confdir, lnam ), source );
                    }
                }
                if ( mfg ) {
                    this.log.info( "%1 manufacturer-specific data for %2 (%3) from %4 ver %5.%6", this,
                        N.manufacturerId, mfg.manufacturer_name, source, mfg.version, mfg.revision );
                    impl.manufacturer[ N.manufacturerId ] = mfg;
                } else {
                    this.log.warn( "%1 no manufacturer-specific data for node %2 mfg %3", this, N.nodeId,
                        N.manufacturerId );
                }
            } catch ( err ) {
                this.log.exception( err );
            }
        }

        /* Identify device type */
        const mfg = impl.manufacturer[ N.manufacturerId ];
        let dt;
        if ( mfg ) {
            this.log.debug( 5, "%1 node %2 looking for manufacturer %3 type %4 id %5", this, N.nodeId, N.manufacturerId,
                N.productType, N.productId );
            let ptyp = N.productType;
            let pid = N.productId;
            let count = 16;
            while ( --count ) {
                if ( mfg.devices[ ptyp ] ) {
                    if ( mfg.devices[ ptyp ][ pid ] ) {
                        dt = mfg.devices[ ptyp ][ pid ];
                    } else {
                        // ??? no... needs more cowbell  // dt = mfg.devices[ ptyp ];
                    }
                    if ( "number" === typeof dt ) {
                        pid = dt;
                        continue;
                    }
                    if ( ! isUndef( dt?.alias ) ) {
                        if ( "number" === typeof dt.alias ) {
                            pid = dt.alias;
                        } else if ( Array.isArray( dt.alias ) ) {
                            if ( dt.alias.length === 1 ) {
                                pid = dt.alias[0];
                            } else {
                                ptyp = dt.alias[0];
                                pid = dt.alias[1];
                            }
                        } else {
                            this.log.err( "%1 invalid alias for manufacturer %2 type %3 id %4: (%5)%6",
                                this, N.manufacturerId, ptyp, pid, typeof dt.alias, dt.alias );
                            break;
                        }
                        this.log.debug( 6, "%1 manufacturer support for %2,%3 aliased to %4,%5", this, N.productType,
                            N.productId, ptyp, pid );
                        continue;
                    }
                    if ( dt ) {
                        /* If what we found is basically description and model only, null it out, we don't really have it. */
                        if ( ! ( dt.entity_class || dt.primary_attribute || dt.capabilities || dt.capability ) ) {
                            this.log.debug( 5, "%1 manufacturer-specific data for type %2 id %3 identity only", this,
                                ptyp, pid );
                            dt = false;
                        } else {
                            this.log.debug( 5, "%1 manufacturer-specific device support, %2", this, dt.description || dt.model );
                            dt.manufacturerId = N.manufacturerId;
                            dt.productType = ptyp;
                            dt.productId = pid;
                        }
                    } else {
                        this.log.debug( 5, "%1 node %2 no manufacturer-specific device support", this, N.nodeId );
                    }
                } else {
                    this.log.debug( 5, "%1 no product type %2 in manufacturer %3 data", this, ptyp, N.manufacturerId );
                }
                break;
            }
        }
        /* Use spread to apply manufacturer-specific device support over generic device support (merge with mfg priority) */
        dt = { ...( this._match_devicetype( N ) || {} ), ...( dt || {} ) };

        lexp.define_var( ctx, 'node', N );

        for ( const endpoint of N.endpoints ) {

            this.log.debug( 5, "%1 handling node %2-%3 with %4 values", this, N.nodeId, endpoint.index,
                endpoint.__values.length );
            lexp.define_var( ctx, 'endpoint', endpoint );
            const entity_id = `${N.nodeId}-${endpoint.index}`;
            const esig = `${impl.version}:${impl.revision}:${mfg?.version||0}:${mfg?.revision||0}`;
            let entity = this.findEntity( entity_id );
            if ( entity ) {
                if ( entity.getAttribute( 'zwave_device.impl_sig' ) !== esig ) {
                    this.log.debug( 5, "%1 clearing stale entity %2; signature mismatch", this, entity );
                    entity.clearCapabilities();
                }
            }
            const ndt = ( 0 === endpoint.index ? dt : this._match_devicetype( endpoint ) ) || {};
            if ( ! entity ) {
                this.log.debug( 6, "%1 creating new entity for %2", this, entity_id );
                entity = this.getEntity( ndt.entity_class || "Entity", entity_id );
                entity.setName(
                    ( N.name || N.label ||  ndt.default_name || N.deviceConfig?.label || N.deviceConfig?.description || `Node ${N.nodeId}` ) +
                    ( endpoint.index > 0 ? ` ${endpoint.index}` : "" ) );

                if ( 0 === endpoint.index ) {
                    this.sendNotice( "Discovered new node {0:q} (#{1}, status {3}). {2}", entity.getName(), N.nodeId,
                        N.ready ? "" : "The node was not ready at the time of discovery; Z-Wave JS may still be interviewing it, or it's unresponsive. The node will not have complete capabilities here until it becomes ready.",
                        NODE_STATUS_TEXT[ N?.status || 0 ] || "unrecognized" );
                }
            }
            entity.deferNotifies( true );
            try {
                entity.markDead( false );

                entity.extendCapability( "zwave_device" );
                entity.setAttribute( "zwave_device.node_id", N.nodeId );
                entity.setAttribute( "zwave_device.endpoint", endpoint.index || 0 );
                entity.setAttribute( 'zwave_device.impl_sig', esig );

                entity.__act = {};

                entity.setType( ndt.entity_class || "Entity" );

                // Ref: https://zwave-js.github.io/node-zwave-js/#/api/node?id=zwavenode-properties
                if ( 0 === endpoint.index ) {
                    entity.setAttribute( "zwave_device.failed", !N.ready );
                    entity.setAttribute( "zwave_device.status", coalesce( N.status ) );
                    entity.setAttribute( "zwave_device.status_text", NODE_STATUS_TEXT[ N.status || 0 ] || "unrecognized" );
                    entity.setAttribute( "zwave_device.generic_class", coalesce( N.deviceClass?.generic?.label ) );
                    entity.setAttribute( "zwave_device.specific_class", coalesce( N.deviceClass?.specific?.label ) );
                    entity.setAttribute( "zwave_device.manufacturer_info",
                        [ N.manufacturerId, N.productType, N.productId ] );
                    entity.setAttribute( "zwave_device.version_info", [ coalesce( N.version ), coalesce( N.firmwareVersion ) ] );
                    entity.setAttribute( "zwave_device.is_listening", !!N.isListening );
                    entity.setAttribute( "zwave_device.is_routing", !!N.isRouting );
                    entity.setAttribute( "zwave_device.is_secure", "boolean" === typeof N.isSecure ? N.isSecure : null );
                    entity.setAttribute( "zwave_device.is_beaming", !!N.isBeaming );
                    entity.setAttribute( "zwave_device.max_data_rate", coalesce( N.maxBaudRate ) );
                }

                for ( let dtcap of ( ndt.capabilities || [] ) ) {
                    entity.extendCapability( dtcap );
                }
                entity.setAttribute( "x_debug.dt", ndt );

                lexp.define_var( ctx, 'entity', entity.extract() );

                const endpointCC = new Set();

                /* ??? For each endpoint, in may be useful to have all of the values in ctx indexed by valueId, or some
                 * other structure with a getValue() function defined? As an example, Meter values may have supportReset
                 * property to tell whether setting the reset property can reset the meter value(s).
                 */
                lexp.define_var( ctx, 'values', endpoint.__valueIndex );

                let values = new Map( Object.entries( endpoint.__valueIndex ) );

                /* Handle values matched to the device type. */
                let handled = new Set();
                for ( const [valueId,value] of values.entries() ) {
                    let matched = false;
                    for ( const map of ( ndt.values || [] ) ) {
                        if ( ! ( map.valueId || map.match ) ) {
                            this.log.err( "%1 invalid mapping %2", this, map );
                            continue;
                        }
                        if ( map.valueId && valueId !== map.valueId ) {
                            continue;
                        }
                        if ( map.match_endpoint ) {
                            if ( ! this._endpoint_match( map, endpoint ) ) {
                                continue;
                            }
                        }
                        if ( map.match ) {
                            if ( isUndef( map.match.commandClass ) ) {
                                this.log.warn( "%1 device class data missing match to commandClass: %2", this, map );
                            }
                            if ( map.unmatched && matched ) {
                                continue;
                            }
                            if ( ! this._value_match( map, value ) ) {
                                continue;
                            }
                        }

                        /* We're doing this! */
                        matched = true;
                        this.log.debug( 5, "%1 handling %2:%3 with device-specific mapping (value %4)", this,
                            endpoint.index, valueId, value.value );
                        lexp.define_var( ctx, 'current_value', value );
                        if ( map.explode ) {
                            // Special naming?
                            const child_id = entity.getID() + '-' + value.propertyKey;
                            let child = this.findEntity( child_id );
                            if ( ! child ) {
                                child = this.getEntity( map.entity_class || "Entity", child_id );
                                child.setName( `${entity.getName()} ${value.metadata?.label || ( value.property + ' ' + value.propertyKey )}` );
                            }
                            child.__act = child.__act || {};
                            child.markDead( false );
                            this._handle_value( child, N, endpoint, valueId, value, map, ctx );
                            if ( map.primary_attribute ) {
                                try {
                                    child.setPrimaryAttribute( map.primary_attribute );
                                } catch ( err ) {
                                    this.log.err( "%1 can't set primary attribute on %2: %3", this, child, err );
                                }
                            }
                            child.deferNotifies( false );
                        } else {
                            this._handle_value( entity, N, endpoint, valueId, value, map, ctx );
                        }
                        handled.add( valueId );
                    }
                }
                /* Remove values handled at this stage */
                for ( let valueId of handled.values() ) {
                    values.delete( valueId );
                }

                /* Handle each command class, looking for remaining values in value-specific handling */
                handled.clear();
                for ( const cc of endpoint.__commandClasses.values() ) {
                    let cmap = coalesce( impl.command_class[ cc ] );
                    if ( false === cmap ) {
                        this.log.debug( 5, "%1 skipping excluded commandClass 0x%3", this, cc.toString(16) );
                        continue;
                    }
                    if ( ! cmap ) {
                        this.log.warn( "%1 no command class mapping for 0x%2; ignored", this, cc.toString(16) );
                        continue;
                    }

                    if ( cmap.device_class ) {
                        if ( cmap.device_class[ N.deviceClass.generic.label ] &&
                            cmap.device_class[ N.deviceClass.generic.label ][ N.deviceClass.specific.label ] ) {
                            cmap = cmap.device_class[ N.deviceClass.generic.label ][ N.deviceClass.specific.label ];
                            this.log.debug( 5, "%1 handling commandClass 0x%2 values (specific %2 support)", this,
                                cc.toString(16), N.deviceClass.specific.label );
                        } else if ( cmap.device_class[ N.deviceClass.specific.label ] ) {
                            cmap = cmap.device_class[ N.deviceClass.specific.label ];
                            this.log.debug( 5, "%1 handling commandClass 0x%2 values (specific %2 support)", this,
                                cc.toString(16), N.deviceClass.specific.label );
                        }
                    } else {
                        this.log.debug( 5, "%1 handling commandClass 0x%2 values (generic support)", this, cc.toString(16) );
                    }

                    if ( cmap.values ) {
                        for ( let [valueId,value] of values.entries() ) {
                            if ( value.commandClass !== cc ) {
                                continue;
                            }
                            let matched = false;
                            for ( let map of cmap.values ) {
                                if ( map.valueId && valueId !== map.valueId ) {
                                    continue;
                                }
                                if ( map.unmatched && matched ) {
                                    continue;
                                }
                                if ( map.match_endpoint ) {
                                    if ( ! this._endpoint_match( map, endpoint ) ) {
                                        continue;
                                    }
                                }
                                if ( map.match ) {
                                    if ( ! this._value_match( map, value ) ) {
                                        continue;
                                    }
                                }

                                /* We go! */
                                matched = true;
                                this.log.debug( 5, "%1 handling %2 with command class mapping", this, valueId );
                                lexp.define_var( ctx, 'current_value', value );
                                if ( map.explode ) {
                                    // Special naming?
                                    const child_id = entity.getID() + '-' + value.propertyKey;
                                    let child = this.findEntity( child_id );
                                    if ( ! child ) {
                                        child = this.getEntity( map.entity_class || "Entity", child_id );
                                        child.setName( `${entity.getName()} ${value.metadata?.label || ( value.property + ' ' + value.propertyKey )}` );
                                    }
                                    child.__act = child.__act || {};
                                    child.markDead( false );
                                    this._handle_value( child, N, endpoint, valueId, value, map, ctx );
                                    if ( map.primary_attribute ) {
                                        try {
                                            child.setPrimaryAttribute( map.primary_attribute );
                                        } catch ( err ) {
                                            this.log.err( "%1 can't set primary attribute on %2: %3", this, child, err );
                                        }
                                    }
                                    child.deferNotifies( false );
                                } else {
                                    this._handle_value( entity, N, endpoint, valueId, value, map, ctx );
                                }
                                handled.add( valueId );
                            }
                        }
                    }

                    for ( const capname of Object.keys( cmap.capabilities || {} ) ) {
                        const cap = cmap.capabilities[ capname ];
                        this.log.debug( 5, "%1 commandClass 0x%2 capability %3", this, cc.toString(16), capname );
                        entity.extendCapability( capname );
                        for ( const attrname of Object.keys( cap.attributes || {} ) ) {
                            const fullname = `${capname}.${attrname}`;
                            const attr = cap.attributes[ attrname ];
                            if ( attr.property ) {
                                attr.valueId = `${coalesce( attr.commandClass, cc )}:${attr.property}:${BLANK(attr.propertyKey)}`;
                                delete attr.commandClass;
                                delete attr.property;
                                delete attr.propertyKey;
                            }
                            if ( attr.valueId ) {
                                /* From different value on this endpoint, any CC */
                                const vv = endpoint.__valueIndex[ attr.valueId ];
                                if ( vv ) {
                                    /* Save mapping for updates */
                                    vv.__endp = vv.__endp || {};
                                    vv.__endp[ fullname ] = { entity: entity.getID(), impl: attr };
                                }
                                handled.add( attr.valueId );
                                this.log.debug( 6, "%1 (CC %5) deriving %2 %3 from %4", this, entity, fullname, attr, cc );
                            }
                            if ( ! attr.update_only || ! entity.hasAttribute( fullname ) ) {
                                this._apply_value( entity, fullname, endpoint, attr, ctx );
                            }
                        }

                        /* And now the actions for this capability, if they haven't already been defined. */
                        for ( const actname of Object.keys( cap.actions || {} ) ) {
                            const key = `${capname}.${actname}`;
                            const ii = cap.actions[ actname ];
                            if ( ! ii ) { /* false or null means not implemented here (or at all) */
                                continue;
                            }
                            if ( ! entity.__act[ key ] ) {
                                if ( "object" === typeof ii ) {
                                    entity.__act[ key ] = {
                                            commandClass: cc,
                                            endpoint: endpoint.index,
                                            impl: ii
                                    };
                                } else {
                                    entity.__act[ key ] = ii;
                                }
                                entity.registerAction( `${capname}.${actname}` );
                            } else {
                                this.log.debug( 5, "%1 entity %2 action %3 already defined by %4", this, entity, key,
                                    entity.__act[ key ] );
                            }
                        }
                    }
                }
                for ( let valueId of handled.values() ) {
                    values.delete( valueId );
                }

                /* Finally, handle remaining values based on generic command class mapping */
                for ( const [valueId,value] of values.entries() ) {
                    this.log.debug( 6, "%1 node %2-%3 valueId %4 default extended (%5 0x%6)", this, N.nodeId,
                        endpoint.index, valueId, value.commandClassName, value.commandClass.toString(16) );
                    this.log.debug( 7, "%1 value (%2)%3", this, typeof value, value );
                }

                /* Once more for all the extended attributes */
                for ( const value of endpoint.__values ) {
                    /* Define value as extended attribute */
                    let fullname = `x_zwave_values.`;
                    fullname += String( value.commandClassName || value.CommandClass ).replace( /[^A-Z0-9_]/ig, "_" );
                    fullname += "_" + String( value.propertyName ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                    if ( "undefined" !== typeof value.propertyKey && null !== value.propertyKey ) {
                        fullname += "_" + String( value.propertyKey ).replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                    }
                    entity.setAttribute( fullname, coalesce( value.value ) );
                    try {
                        entity.setAttributeMeta( fullname, 'zwave_property', value.property );
                    } catch ( err ) {
                        this.log.err( "%1 error setting metadata for %2 %3: %4", this, entity, fullname, err );
                    }
                }

                entity.setAttribute( 'zwave_device.capabilities', Array.from( endpoint.__commandClasses ) );

                if ( ndt.primary_attribute ) {
                    try {
                        entity.setPrimaryAttribute( ndt.primary_attribute );
                    } catch ( err ) {
                        this.log.err( "%1 can't set primary attribute on %2: %3", this, entity, err );
                    }
                }

            } catch ( err ) {
                this.log.exception( err );
            } finally {
                entity.deferNotifies( false );
            }

            try {
                this.log.info( "%1 configuring node %2 endpoint %3 (entity %4)", this, N.nodeId, endpoint.index, entity.getID() );
                this._configure_endpoint( entity, endpoint );
            } catch ( err ) {
                this.log.err( "%1 failed %2-%3 configuration: %4", this, N.nodeId, endpoint.index, err );
                this.log.exception( err );
            }

            this.log.debug( 5, "%1 node %2-%3 setup complete", this, N.nodeId, endpoint.index );
        }

        this.log.debug( 5, "%1 node %2 registration complete", this, N.nodeId );
        return N;
    }

   _handle_value( entity, node, endpoint, valueId, value, map, ctx ) {

       if ( ! isUndef( map.withCC ) ) {
            if ( ! endpoint.__commandClasses.has( map.withCC ) ) {
                return;
            }
        }
        if ( ! isUndef( map.withoutCC ) ) {
            if ( endpoint.__commandClasses.has( map.withoutCC ) ) {
                return;
            }
        }

        if ( ! map.capabilities && map.capability ) {
            this.log.warn( "%1 node %2-%3 translating single-capability mapping in %4 for %5", this, node.nodeId,
                endpoint.index, map, valueId );
            map.capabilities = { [map.capability]: { attributes: map.attributes || {}, actions: map.actions || {} } };
            delete map.capability;
            delete map.attributes;
            delete map.actions;
        }

        for ( let capname of Object.keys( map.capabilities || {} ) ) {
            const cap = map.capabilities[ capname ];

            this.log.debug( 7, "%1 node %2-%3 CC %4 (0x%5) capability %6", this, node.nodeId,
                endpoint.index, value.commandClassName, value.commandClass.toString(16), capname );

            entity.extendCapability( capname );

            /* Local actions defined? */
            for ( let actname of Object.keys( cap.actions || {} ) ) {
                const key = `${capname}.${actname}`;
                const ii = cap.actions[ actname ];
                if ( ! ii ) {  /* not implemented here, or at all */
                    continue;
                }
                if ( ! entity.__act[ key ] ) {
                    if ( "object" === typeof ii ) {
                        /* Overrides in implementation are handled in performAction() */
                        entity.__act[ key ] = {
                            endpoint: endpoint.index,
                            commandClass: value.commandClass,
                            property: value.property,
                            propertyKey: value.propertyKey,
                            impl: ii
                        };
                    } else {
                        entity.__act[ key ] = ii;
                    }
                    this.log.debug( 5, "%1 node %2-%3 value-based action definition %4.%5 for %6", this, node.nodeId,
                        endpoint.index, capname, actname, valueId );
                    entity.registerAction( key );
                } else {
                    this.log.debug( 5, "%1 entity %2 action %3 already defined by %4", this, entity, key,
                        entity.__act[ key ] );
                }
            }

            /* Now handle the attributes cc */
            for ( let attrname of Object.keys( cap.attributes || {} ) ) {
                const fullname = `${capname}.${attrname}`;
                let attr = cap.attributes[ attrname ];
                if ( isUndef( attr.valueId ) ) {
                    /* valueId not specified in implementation, so derived from current value; shallow-copy impl and
                     * set valueId for this value.
                     */
                    attr = { ...attr };
                    attr.valueId = valueId;
                }

                /* Save mapping for updates */
                value.__endp = value.__endp || {};
                value.__endp[ fullname ] = { entity: entity.getID(), impl: attr };
                this.log.debug( 6, "%1 deriving %2 %3 from %4", this, entity, fullname, attr );

                if ( ! attr.update_only || ! entity.hasAttribute( fullname ) ) {
                    this._apply_value( entity, fullname, endpoint, attr, ctx );
                }
            }
        }
    }

    _remove_node( nodeId ) {
        this.log.info( "%1 removing node %2", this, nodeId );
        delete this.nodes[ nodeId ];

        /* Remove entities for this node */
        for ( let entity of Object.values( this.getEntities() ) ) {
            if ( entity.hasCapability( 'zwave_device' ) && nodeId === entity.getAttribute( 'zwave_device.node_id' ) ) {
                this.log.debug( 5, "%1 removing entity %2 for node %3", this, entity, nodeId );
                try {
                    this.removeEntity( entity );
                } catch ( err ) {
                    this.log.err( "%1 failed to remove entity %2: %3", this, entity, err );
                    this.log.exception( err );
                }
            }
        }
    }

    _configure_endpoint( entity, endpoint ) {
        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        this.log.debug( 5, "%1 endpoint config entity %2 commandClasses=%3", this, entity, Array.from( endpoint.__commandClasses ) );
        // ??? Do we need to enforce group 1 to controller (lifeline)?
        if ( endpoint.__commandClasses.has( 0x2b ) ) {
            /* Endpoint has Scene Activation CC */
            if ( endpoint.__commandClasses.has( 0x2d ) ) {
                /* Ref: 4.85 -- Scene Controller Configuration is used to configure nodes launching scenes using
                 * their association groups. So for any sceneId value in the command class that doesn't have a value,
                 * set an association to the controller where the groupId = propertyKey (often same as button number).
                 */
                this.log.debug( 5, "%1 checking/setting endpoint configuration for %2 Scene Activation CC", this, entity );
                for ( let value of endpoint.__values ) {
                    if ( 0x2d === value.commandClass && 'sceneId' === value.property && isUndef( value.value ) ) {
                        this.log.info( "%1 entity %2 adding association for unconfigured sceneId group #%3",
                            this, entity, value.propertyKey );
                        this._add_association( nodeId, endpoint.index, value.propertyKey, [ 1 ] );
                    }
                }
            } else if ( endpoint.__commandClasses.has( 0x2c ) ) {
                /* Ref: 4.83 -- Scene Actuator Configuration is used for launching scenes in a number of actuator
                 * nodes (e.g. another scene controlling unit, multi-level switch, binary switch, etc.
                 */
                this.log.debug( 5, "%1 checking/setting endpoint configuration for %2 Scene Actuator CC", this, entity );
                for ( let value of endpoint.__values ) {
                    if ( 0x2c === value.commandClass && 'level' === value.property && isUndef( value.value ) ) {
                        this.log.info( "%1 entity %2 adding Scene Actuator level for unconfigured value #%3", this,
                            entity, value.propertyKey );
                        this._set_endpoint_value( endpoint, value.commandClass, value.property, value.propertyKey,
                            value.propertyKey );
                    }
                }
            }
        }
    }

    async _add_association( sourceNodeId, sourceEndpoint, groupId, targets ) {
        let req = {
            command: 'controller.add_associations',
            nodeId: sourceNodeId,
            group: groupId,
            associations: []
        };
        const N = await this.nodes[ sourceNodeId ];
        const endp = N.endpoints[ sourceEndpoint ];
        if ( endp.__commandClasses.has( 0x8e ) ) {
            /* Has multi-channel association; send with endpoints */
            req.endpoint = sourceEndpoint;
            for ( let a of targets ) {
                req.associations.push( "number" === typeof a ? { nodeId: a, endpoint: 0 } : a );
            }
        } else {
            /* No multi-channel associations; set simple */
            for ( let a of targets ) {
                req.associations.push( "number" === typeof a ? { nodeId: a } : { nodeId: a.nodeId } );
            }
        }
        return this._send_tracked_message( req ).catch( err => {
            this.log.err( "%1 failed to set associations on %2 group %3: %4", this, entity, groupId, err );
            this.log.err( "%1 request was %2", this, req );
        });
    }

    /**
    *  Refreshes a node (interview). Returns a promise that resolves when the refresh is done.
    */
    _refresh_node_info( nodeId ) {
        return this._send_tracked_message( { command: "node.refresh_info", nodeId: nodeId } );
    }

    /**
     * Set a value on a node/endpoint.
     */
    async _set_value( nodeId, endpoint, commandClass, property, propertyKey, newval ) {
        this.log.debug( 5, "%1 _set_value(%2,%3,%4,%5,%6,%7)", this, nodeId, endpoint, commandClass, property,
            propertyKey, newval );
        const N = await this.nodes[ nodeId ];
        if ( ! N ) {
            return Promise.reject( new ReferenceError( `Unknown node ${nodeId}` ) );
        }
        const endp = N.endpoints[ endpoint || 0 ];
        if ( ! endp ) {
            return Promise.reject( new ReferenceError( `Unknown endpoint ${nodeId}:${endpoint||0}` ) );
        }

        return this._set_endpoint_value( endp, commandClass, property, propertyKey, newval );
    }

    async _set_endpoint_value( endp, commandClass, property, propertyKey, newval ) {
        /* Command class name -> number */
        if ( "string" === typeof commandClass ) {
            if ( isNaN( commandClass ) ) {
                commandClass = CC.has( commandClass ) ? CC.get( commandClass ) : commandClass;
            } else {
                commandClass = parseInt( commandClass );  /* Number as string, convert */
            }
        }

        const vid = commandClass + ":" + property + ":" + BLANK( propertyKey );
        const v = endp.__valueIndex[ vid ];
        if ( isUndef( v ) ) {
            let e = new ReferenceError( `No such value on endpoint ${endp.nodeId}-${endp.index} ${vid}` );
            e.valueId = vid;
            e.node = endp.nodeId;
            e.endpoint = endp.index || 0;
            return Promise.reject( e );
        }
        if ( false === v.writeable ) {
            return Promise.reject( new Error( `Value is not writeable` ) );
        }

        /* Sanitize data */
        if ( "number" === v.metadata?.type ) {
            newval = parseFloat( newval );
            if ( isNaN( newval ) ) {
                return Promise.reject( new TypeError( `Invalid value type, expecting ${v.metadata.type}` ) );
            }
            if ( ( undefined !== v.metadata.min && newval < v.metadata.min ) ||
                 ( undefined !== v.metadata.max && newval > v.metadata.max ) ) {
                return Promise.reject( new RangeError() );
            }
            if ( undefined !== v.metadata.steps ) {
                newval = Math.floor( newval / v.metadata.steps + 0.5 ) * v.metadata.steps;
            }
        } else if ( "boolean" === v.metadata?.type ) {
            if ( "string" === typeof newval ) {
                newval = TRUE_RE.test( newval );
            } else {
                newval = !! newval;
            }
        }

        /* Use the identiers from the matched value object (helps with types; some values have numeric properties */
        let p = { command: "node.set_value", nodeId: endp.nodeId };
        p.endpoint = v.endpoint || 0;
        p.valueId = { commandClass: v.commandClass, property: v.property };
        if ( "undefined" !== typeof v.propertyKey && null !== v.propertyKey ) {
            p.valueId.propertyKey = v.propertyKey;
        }
        p.value = newval;
        this.log.debug( 5, "%1 setting value %2", this, p );
        return this._send_tracked_message( p );
    }

    /**
    *  Refreshes a node. Returns a promise that resolves when the refresh is done.
    */
    action_zwave_device_refresh( entity ) {
        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        return this._send_tracked_message( { command: "node.refresh_values", nodeId: nodeId } );
    }

    /* Deprecated form */
    action_zwave_device_poll( entity ) {
        this.log.warn( "%1 using zwave_device.poll is deprecated; use zwave_device.refresh", this );
        return this.action_zwave_device_refresh( entity );
    }

    action_zwave_device_interview( entity ) {
        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        return this._refresh_node_info( nodeId );
    }

    /* ??? Not permanent? Docs say need to invoke Naming and Location CC for permanent change.
     * endpoint.invoke_cc_api { commandClass:, methodName: args: }
     */
    action_zwave_device_set_name( entity, params ) {
        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        const nom = params.name || `Node ${nodeId}`;
        entity.setName( nom ); // ??? Do only on success?
        return this._send_tracked_message( { command: "node.set_name", nodeId: nodeId, args: [ nom ] } );
        // ??? Handle NodeNamingAndLocationCC?
    }

    action_zwave_device_set_location( entity, params ) {
        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        const loc = params.location || null;
        return this._send_tracked_message( { command: "node.set_location", nodeId: nodeId, args: [ loc ] } );
        // ??? Handle NodeNamingAndLocationCC?
    }

    /* Also Interview, ping, firmwareUpdate?, ... */

    action_zwave_device_set_value( entity, params ) {
        this.log.debug( 5, "%1 zwave_device_set_value(%2, %3)", this, entity, params );
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        return this._set_value( nodeId,
            coalesce( params.endpoint, entity.getAttribute( `zwave_device.endpoint` ) ),
            params.commandClass,
            params.property,
            "" !== BLANK( params.propertyKey ) ? propertyKey : undefined,
            params.value
        );
    }

    action_zwave_device_reconfigure( entity, params ) {
        return Promise.reject( new Error( "Action reconfigure not implemented" ) );
    }

    async action_zwave_device_set_config( entity, params ) {
        const CC_CONFIGURATION = 112;
        const mask_size = [ 0xff, 0xffff, 0, 0xffffffff ];
        if ( ! mask_size[ params.size - 1 ] ) {
            return Promise.reject( new TypeError( `Size must be 1, 2 or 4` ) );
        }
        let value = parseInt( params.value );
        if ( isNaN( value ) ) {
            return Promise.reject( new TypeError( `Value must be integer` ) );
        }
        let mask = parseInt( params.bitmask || mask_size[ params.size - 1 ] );
        if ( isNaN( mask ) ) {
            return Promise.reject( new TypeError( `Mask must be integer` ) );
        }

        const nodeId = entity.getAttribute( 'zwave_device.node_id' );
        const endp = entity.getAttribute( 'zwave_device.endpoint', 0 );
        const N = await this.nodes[ nodeId ];
        const endpoint = N?.endpoints[ endp ];
        const vid = `${CC_CONFIGURATION}:${params.parameter}:`;
        let v = endpoint?.__valueIndex[ vid ];
        if ( ! v ) {
            /* Try search by property name; property key does not apply here. */
            const pn = params.parameter.toLowerCase().replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
            for ( let vk in ( endpoint?.__valueIndex || {} ) ) {
                const vv = endpoint.__valueIndex[ vk ];
                if ( CC_CONFIGURATION === vv.commandClass ) {
                    const vpn = vv.propertyName.toLowerCase().replace( /[^A-Z0-9_]+/ig, "_" ).replace( /_+$/, "" );
                    if ( vpn === pn ) {
                        v = vv;
                        break;
                    }
                }
            }
        }
        if ( ! v ) {
            return Promise.reject( new ReferenceError( `Configuration parameter not defined` ) );
        }
        /* Apply only those bits of the given value to the final value. */
        if ( params.bitmask ) {
            this.log.debug( 5, "%1 set_config current value %2b mask %3b new value %4b masked %5b", this,
                v.value.toString(2), mask.toString(2), value.toString(2), (value & mask).toString(2) );
        } else {
            this.log.debug( 5, "%1 set_config value %2", this, value );
        }
        value = ( v.value & ( ~mask ) ) | ( value & mask );
        return this._set_endpoint_value( endpoint, CC_CONFIGURATION, v.property, undefined, value );
    }

    action_zwave_device_send_data( entity, params ) {
        return Promise.reject( new Error( "Action reconfigure not implemented" ) );
    }

    action_zwave_device_heal( entity, params ) {
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        let req = { command: "controller.heal_node", nodeId: nodeId };
        this.log.debug( 5, "%1 action_zwave_device_heal sending %2", this, req );
        return this._send_tracked_message( req );
    }

    action_zwave_remove_failed( entity, params ) {
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        let req = { command: "controller.remove_failed_node", nodeId: nodeId };
        this.log.debug( 5, "%1 action_zwave_remove_failed sending %2", this, req );
        return this._send_tracked_message( req );
    }

    action_zwave_device_replace_failed( entity, params ) {
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        let req = { command: "controller.replace_failed_node", nodeId: nodeId };
        this.log.debug( 5, "%1 action_zwave_device_replace_failed sending %2", this, req );
        return this._send_tracked_message( req );
    }

    action_zwave_device_test_powerlevel( entity, params ) {
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        let req = { command: "controller.test_powerlevel", nodeId: nodeId };
        this.log.debug( 5, "%1 action_zwave_device_test_powerlevel sending %2", this, req );
        return this._send_tracked_message( req ).then( ret => {
            this.log.notice( "%1 response from power level test on %2 (node %3): %4", this, entity, nodeId, ret );
        });
    }

    action_zwave_device_reset_meters( entity, params ) {
        /* Set value 50:reset: to true */
        return this._set_value( entity.getAttribute( 'zwave_device.node_id' ), 0, 50 /* Meters */, 'reset', undefined,
            true );
    }

    action_zwave_network_start_include( entity, params ) {
        return this._send_tracked_message( { command: "controller.begin_inclusion" } );
    }

    action_zwave_network_stop_include( entity, params ) {
        return this._send_tracked_message( { command: "controller.stop_inclusion" } );
    }

    action_zwave_network_start_exclude( entity, params ) {
        return this._send_tracked_message( { command: "controller.begin_exclusion" } );
    }

    action_zwave_network_stop_exclude( entity, params ) {
        return this._send_tracked_message( { command: "controller.stop_exclusion" } );
    }

    action_zwave_network_start_heal( entity, params ) {
        return this._send_tracked_message( { command: "controller.begin_healing_network" } );
    }

    action_zwave_network_stop_heal( entity, params ) {
        return this._send_tracked_message( { command: "controller.stop_healing_network" } );
    }

    action_sys_system_restart( entity, params ) {
        this.ws_close();
        return Promise.resolve();
    }

    action_zwave_device_debug( entity, params ) {
        if ( ! this.config.toggledbits ) {
            return Promise.reject( "No implementation" );
        }
        const nodeId = entity.getAttribute( `zwave_device.node_id` );
        // let req = { command: "endpoint.invoke_cc_api", nodeId: nodeId };
        // req.endpoint = 0;
        // req.commandClass = 0x25; // BinarySwitchCC
        // req.methodName = "get_associations";
        // req.args = [
        // ];
        let req = { command: "driver.soft_reset", nodeId: nodeId };
        return this._send_tracked_message( req ).then( ret => {
            this.log.always( "%1 zwave_device_debug %2 returned %3", this, req, ret );
            return ret;
        });
    }

    /* Capability "button" default actions */

    action_button_press( entity, params ) {
        const p = ['unknown','single','double','triple','4','5'];
        entity.setAttributes( {
            'button.state': p[ params.count || 1 ] || 'unknown',
            'button.since': Date.now()
        });
        return Promise.resolve();
    }

    action_button_single( entity, params ) {
        return this.action_button_press( entity, { count: 1 } );
    }

    action_button_double( entity, params ) {
        return this.action_button_press( entity, { count: 2 } );
    }

    action_button_hold( entity, params ) {
        entity.setAttributes( {
            'button.state': 'hold',
            'button.since': Date.now()
        });
        return Promise.resolve();
    }

    action_button_release( entity, params ) {
        entity.setAttributes( {
            'button.state': 'release',
            'button.since': Date.now()
        });
        return Promise.resolve();
    }

    /* The big performer */

    performOnEntity( entity, action, params ) {
        this.log.info( "%1 performing %2 on %3 with %4", this, action, entity, params );
        const def = entity.__act ? entity.__act[ action ] : undefined;
        if ( false === def ) {
            /* Explicit false means not implemented */
            this.log.err( "%1 entity %2 action %3 not implemented/available for this entity", this, entity, action );
        } else if ( ! def ) {
            /* null or undefined -- try default action */
            this.log.debug( 5, "%1 no implementation mapped; attempting default", this );
            return super.performOnEntity( entity, action, params );
        } else if ( true === def ) {
            this.log.debug( 5, "%1 no implementation mapped; attempting default", this );
            return super.performOnEntity( entity, action, params );
        } else if ( "string" === def ) {
            /* Names a function */
            try {
                this[ def ]( entity, action, params );
            } catch ( err ) {
                this.log.err( "%1 entity %2 action %3 func %4 failed: %5", this, entity, action, def, err );
                this.log.exception( err );
                return Promise.reject( err );
            }
        }
        /* Something else */
        this.log.debug( 5, "%1 def %2", this, def );
        if ( ! Array.isArray( def.impl ) ) {
            def.impl = [ def.impl ];
        }
        let p = [];
        for ( const step of def.impl ) {
            if ( "set_value" ===  ( step.command || "set_value" ) ) {
                const nodeId = entity.getAttribute( "zwave_device.node_id" );
                let vid = step.valueId || "";
                if ( "" !== vid ) {
                    /** The "+" form is used in command class-driven actions (command_class section of zwavejs_devices) as
                     *  a shortcut to indicate the subject command class. Substitute that in, along with the endpoint index.
                     *  +targetValue: => endpoint:commandClass:targetValue: (length=4 when split)
                     */
                    if ( vid.startsWith( '+' ) && vid.indexOf( ':' ) < 0 ) {
                        this.log.warn( "Action %1 impl value '%2' deprecated form; correcting", action, vid );
                        this.log.info( "Implementation: %1", step );
                        vid += ':';
                    }
                    vid = vid.replace( /^[+]/,
                        String( coalesce( step.endpoint, def.endpoint, entity.getAttribute( "zwave_device.endpoint" ) ) ) +
                        ':' +
                        String( coalesce( step.commandClass, def.commandClass ) ) +
                        ':'
                    );
                    vid = vid.split( /:/ );
                    if ( vid.length < 4 ) {
                        vid.unshift(
                            coalesce( step.endpoint, def.endpoint, entity.getAttribute( "zwave_device.endpoint" ) )
                        );
                    }
                } else {
                    vid = [
                        coalesce( step.endpoint, def.endpoint, entity.getAttribute( "zwave_device.endpoint" ) ),
                        coalesce( step.commandClass, def.commandClass ),
                        coalesce( step.property, def.property ),
                        coalesce( step.propertyKey, def.propertyKey )
                    ];
                }
                const endpoint = parseInt( vid[0] );
                const commandClass = parseInt( vid[1] );
                const property = vid[2] || "value";
                const propertyKey = "" !== ( vid[3] || "" ) ? vid[3] : undefined;

                /* Fix up value */
                let value = step.value;
                let newval = null;
                if ( null !== value && "object" === typeof value ) {
                    if ( ! isUndef( value.value ) ) {
                        newval = value.value;
                    } else if ( ! isUndef( value.parameter ) ) {
                        newval = coalesce( params[ value.parameter ] );
                    }
                    if ( value.expr ) {
                        if ( isUndef( expr_ce[ value.expr ] ) ) {
                            try {
                                expr_ce[ value.expr ] = lexp.compile( value.expr );
                            } catch ( err ) {
                                this.log.err( "%1 %4 failed compile of %2: %3", this, value.expr, err, action );
                                expr_ce[ value.expr ] = false;  /* negaive, not undefined */
                            }
                        }
                        if ( expr_ce[ value.expr ] ) {
                            let ctx = lexp.get_context( { value: newval, system: this.getSystemEntity().extract(),
                                entity: entity.extract(), parameters: params } );
                            try {
                                newval = lexp.run( expr_ce[ value.expr ], ctx );
                            } catch ( err ) {
                                this.log.err( "%1 %4 failed eval of %2: %3", this, value.expr, err, action );
                                newval = null;
                            }
                        }
                    }
                    if ( "object" === typeof( value.map ) ) {
                        newval = coalesce( value.map[ newval ], value.map_default, newval );
                    }
                    if ( null === newval && ! isUndef( value.default ) ) {
                        newval = value.default;
                    }
                } else {
                    newval = value;
                }

                this.log.debug( 5, '%1 action %2 set endpoint %3 value %4:%5:%6 = %7', this, action, endpoint,
                    commandClass, property, propertyKey, newval );

                p.push( ( resolve, reject ) => {
                    this._set_value( nodeId, endpoint, commandClass, property, propertyKey, newval ).then( data => {
                        resolve( data );
                    }).catch( err => {
                        reject( err );
                    });
                });
            } else {
                return Promise.reject(
                    new Error( `Unrecognized action command ${String(step.command)} for ${action} on ${entity}` )
                );
            }
        }
        return util.sequencePromises( p, true );
    }

    _poll_resched() {
        if ( this.poll_queue.length > 0 ) {
            this.poll_queue.sort( ( a, b ) => a.due - b.due );
            this.poll_timer.delayms( Math.max( ( this.config.pollfrequency || 10 ) * 1000,
                this.poll_queue[0].due - Date.now() ) );
            this.log.debug( 5, "%1 next poll activity %2", this, this.poll_timer.expires() );
        } else {
            this.log.debug( 5, "%1 no polling activity scheduled; poller stopped", this );
        }
    }

    async _poll_handler( ) {
        /* Poll queue is sorted, so first item is eligible... maybe. */
        const now = Date.now();
        let ix = 0;
        while ( ix < this.poll_queue.length ) {
            let p = this.poll_queue[ ix++ ];
            if ( p.due > now ) {
                /* Everything that follows is after current time. */
                this._poll_resched();
                break;
            }
            this.log.debug( 5, "%1 node %2 eligible for polling", this, p.node );
            const entity = this.findEntity( `${p.node}-${p.endpoint||0}` );
            const stat = entity.getAttribute( 'zwave_device.status_text' );
            if ( 'dead' === stat ) {
                this.log.debug( 5, "%1 skipping poll of node %2, node is dead", this, p.node );
                continue;
            } else if ( 'asleep' === stat ) {
                this.log.debug( 5, "%1 skipping poll of node %2, node is asleep", this, p.node );
                continue;
            }
            try {
                let px;
                if ( p.command_class ) {
                    this.log.debug( 5, "%1 polling node %2 command class 0x%3 (%4)", this, p.node,
                        p.command_class.toString(16), CC.get( p.command_class ) );
                    px = this._send_tracked_message({
                        command: "node.refresh_cc_values",
                        nodeId: p.node,
                        commandClass: p.command_class
                    });
                } else {
                    this.log.debug( 5, "%1 polling node %2", this, p.node );
                    px = this.action_zwave_device_refresh( entity, {} );
                }
                px.then( data => {
                    const realnow = Date.now();
                    p.errors = 0;
                    p.due = realnow + p.interval;
                    this.log.debug( 5, "%1 poll node %2 succeeded, next at %3", this, p.node, p.due );
                    entity.lastpoll = realnow;
                    entity.setAttributes({
                        'zwave_device.consecutive_poll_fails': 0,
                    });
                    this._poll_resched();
                    return data;
                }).catch( () => {
                    ++p.errors;
                    this.log.notice( "%1 poll node %2 fail #%3", this, p.node, p.errors );
                    entity.lastpoll = Date.now();
                    entity.setAttributes( {
                        'zwave_device.consecutive_poll_fails': p.errors,
                    });
                    /* Move this entry to the end of the queue for now. Don't re-sort the queue. Just arm for the next. */
                    this.poll_queue.push( this.poll_queue.splice( ix, 1 ) );
                    this.poll_timer.delayms( this.config.poll_frequency );
                });
            } catch ( err ) {
                this.log.exception( err );
                this.poll_timer.delayms( this.config.poll_frequency );
            }
            break;
        }
    }

    _poll_begin() {
        this.poll_queue = [];
        this.log.debug( 5, "%1 _poll_begin have %2 listed nodes", this, ( this.config.poll_nodes || [] ).length );
        ( this.config.poll_nodes || [] ).forEach( cf => {
            this.log.debug( 6, "poll %1", cf );
            let entry = {};
            let entity;
            if ( cf.entity ) {
                entity = this.getNamedEntity( cf.entity );
            } else if ( isNaN( cf.node ) || ! this.nodes[ cf.node ] ) {
                this.log.err( "%1 (poller) node %2 invalid; ignoring", this, cf.node );
                return;
            } else {
                entity = this.findEntity( `${cf.node}-${cf.endpoint||0}` );
                if ( ! entity ) {
                    this.log.warn( "%1 (poller) entity for node %2 does not exist; ignoring", this, entry.node );
                    return;
                }
            }
            entry.node = entity.getAttribute( 'zwave_device.node_id' );
            entry.endpoint = entity.getAttribute( 'zwave_device.endpoint' );
            entry.interval = coalesce( parseInt( cf.interval ), this.config.poll_interval ) * 1000;
            entry.due = coalesce( entity.lastpoll, Date.now() ) + entry.interval;
            if ( cf.command_class ) {
                if ( "string" === typeof cf.command_class && CC.has( cf.command_class ) ) {
                    entry.command_class = CC.get( cf.command_class );
                } else if ( "number" === typeof cf.command_class && CC.has( cf.command_class ) ) {
                    entry.command_class = cf.command_class;
                } else {
                    this.log.warn( "%1 (poller) ignoring unrecognized command class %2", this, cf.command_class );
                }
            }
            this.poll_queue.push( entry );
            this.log.debug( 6, "%1 enqueued polling entry %2", this, entry );
        });

        this._poll_resched();
    }
};
